from ctypes import cdll, CDLL, RTLD_GLOBAL
from ctypes import POINTER, byref, c_int, c_int64, c_int32, c_bool, c_char_p, c_double, c_void_p, CFUNCTYPE, py_object, cast, byref, Structure, string_at
import ctypes
import traceback
'''
CDLL(MPI.__file__, mode=RTLD_GLOBAL) is a workaround for a few MPICH bugs, including 
the bug with non-working MPI_IN_PLACE and 2-stage ELPA solver
https://bitbucket.org/mpi4py/mpi4py/issues/162/mpi4py-initialization-breaks-fortran
https://lists.mpich.org/pipermail/discuss/2020-July/006018.html
https://github.com/pmodels/mpich/issues/4130
'''
from mpi4py import MPI
CDLL(MPI.__file__, mode=RTLD_GLOBAL)

import numpy as np
from numpy.ctypeslib import ndpointer

import sys
from pathlib import Path
import os, shutil
from ase import units
from scalapack4py import ScaLAPACK4py
from scalapack4py.blacsdesc import ctypes2ndarray
from scipy.linalg import lapack

class ASI_matrix_descr(Structure):
    _fields_ = [("matrix_type", c_int),
                ("storage_type", c_int)]

dmhs_callback = CFUNCTYPE(None, c_void_p, c_int, c_int, POINTER(c_int), POINTER(c_double), POINTER(ASI_matrix_descr))  # void(*)(void *aux_ptr, int iK, int iS, int *blacs_descr, void *blacs_data, ASI_matrix_descr_t *matrix_descr)
set_dmhs_callback = CFUNCTYPE(c_int, c_void_p, c_int, c_int, POINTER(c_int), POINTER(c_double), POINTER(ASI_matrix_descr))  # void(*)(void *aux_ptr, int iK, int iS, int *blacs_descr, void *blacs_data, ASI_matrix_descr_t *matrix_descr)
esp_callback = CFUNCTYPE(None, c_void_p, c_int, POINTER(c_double), POINTER(c_double), POINTER(c_double))   # void(*)(void *aux_ptr, int n, const double *coords, double *potential, double *potential_grad)

class ASI_basis_func_descr_generic(Structure):
  '''
     Generic basis function descriptior.
     Use basis_func_descr_type field to find actual structure type
     in basis_func_descr_ptr_types dict
  '''
  _fields_ = [("descr_size",c_int), ("basis_func_descr_type",c_int),]

class ASI_basis_func_descr_generic_local(Structure):
  _fields_ = [("descr_size",c_int), ("basis_func_descr_type",c_int), ("atom_index",c_int),]

class ASI_basis_func_descr_aims(Structure):
    _fields_ = [("descr_size",c_int), ("basis_func_descr_type",c_int), \
    ("atom_index",c_int),
    ("func_type_ptr", c_char_p),
    ("func_type_len", c_int),
    ("n_qn",c_int),
    ("l_qn",c_int),
    ("m_qn",c_int)]
    
    @property
    def func_type(self):
      return string_at(self.func_type_ptr,self.func_type_len).decode('utf-8')

basis_func_descr_types = {101:ASI_basis_func_descr_generic_local, 102:ASI_basis_func_descr_aims}


def triang2herm_inplace(X, uplo):
  indices_func = {"L":np.triu_indices, "U":np.tril_indices}[uplo]
  diag_shift = {"L":1, "U":-1}[uplo]
  i_upper = indices_func(X.shape[0], diag_shift)
  X[i_upper] = X.T[i_upper].conj()

def triang_packed2full_hermit(data, n, is_real, uplo):
  lapack_triang2full = lapack.dtpttr if is_real else lapack.ztpttr
  n_triang = (n + 1)*n//2
  data_packed_shape = (n_triang, ) if is_real else (n_triang*2, ) # 
  data_packed = np.ctypeslib.as_array(data, shape=data_packed_shape)
  if not is_real:
    data_packed = data_packed.view(np.complex128).reshape((n_triang,))

  data, info = lapack_triang2full(n, data_packed, uplo=uplo)
  triang2herm_inplace(data, uplo)
  return data
  
def HS_to_DM(H, S, nelectrons=None, fermi_level=None):
  '''
    Compute density matrix from Hamiltonian (H) and overlap (S) matrices
    via generalized eigenproblem solving. No smearing is used. Full spin-paired 
    occupancy is assumed.

    Parameters
    ----------
    H : Hamiltonian matrix
    S : Overlap matrix
    nelectrons : number of electrons 
    fermi_level : Fermi level to define number of electrons

    Returns
    -------
    DM : ndarray[n, n]
      Density matrix
  '''
  assert nelectrons is None or fermi_level is None
  from scipy.linalg import eigh
  w,v = eigh(H,S)

  occupancy_number = 2 # number of electrons per orbital - full spin-paired

  if nelectrons is None:
    if fermi_level is None:
      fermi_level = 0.0
    n_occ = np.sum(w < fermi_level) # number of occupied orbitals by energy
  else:
    n_occ = nelectrons // occupancy_number
    assert  nelectrons % occupancy_number == 0
  
  DM = v[:,:n_occ] @ (np.eye(n_occ) * occupancy_number )@ v[:,:n_occ].conj().T

  return DM

def matrix_asi_to_numpy(asi, descr, data, matrix_descr_ptr):
    #if (matrix_descr_ptr.contents.storage_type==0):
  if (matrix_descr_ptr.contents.storage_type not in {1,2}): # temporary workaround in aims implementation
    data = asi.scalapack.gather_numpy(descr, data, asi.matrix_shape)
  elif (matrix_descr_ptr.contents.storage_type in {1,2}): # ASI_STORAGE_TYPE_TRIL,ASI_STORAGE_TYPE_TRIU
    assert not descr, "default_saving_callback supports only dense full ScaLAPACK arrays"
    assert matrix_descr_ptr.contents.matrix_type == 1, "Triangular packed storage is supported only for hermitian matrices"
    uplo = {1:'L',2:'U'}[matrix_descr_ptr.contents.storage_type]
    data = triang_packed2full_hermit(data, asi.n_basis, asi.is_hamiltonian_real, uplo)
  else:
    assert False, f"Unsupported storage_type {matrix_descr_ptr.contents.storage_type}"
  return data

def default_saving_callback(aux, iK, iS, descr, data, matrix_descr_ptr):
  '''
     In this implementation of a callback the performace is sacrificed for universality
  '''
  try:
    label = 'no-label'
    asi, storage_dict, cnt_dict, keep_history, label = cast(aux, py_object).value
    data = matrix_asi_to_numpy(asi, descr, data, matrix_descr_ptr)

    if data is not None:
      assert len(data.shape) == 2
      data = data.copy()
      if asi.implementation == "DFTB+":
        triang2herm_inplace(data, 'L')
      if keep_history:
        if (iK, iS) not in storage_dict:
          storage_dict[(iK, iS)] = []
        storage_dict[(iK, iS)].append(data)
      else:
        storage_dict[(iK, iS)] = data
        triang2herm_inplace(storage_dict[(iK, iS)], 'U')
    cnt_dict[(iK, iS)] = cnt_dict.get((iK, iS), 0) + 1
  except Exception as eee:
    print (f"Something happened in ASI default_saving_callback {label}: {eee}\nAborting...")
    print ("T:", traceback.format_exc())
    MPI.COMM_WORLD.Abort(1)

def default_loading_callback(aux, iK, iS, descr, data, matrix_descr_ptr):
  try:
    label = 'no-label'
    asi, storage_dict, label = cast(aux, py_object).value
    m = storage_dict[(iK, iS)] if asi.scalapack.is_root(descr) else None
    assert m is None or (asi.n_basis == m.shape[0]) and (asi.n_basis == m.shape[1]), \
                     f"m.shape=={m.shape} != asi.n_basis=={asi.n_basis}"
    asi.scalapack.scatter_numpy(m, descr, data)
    return 1 # signal that  matrix has been loaded
  except Exception as eee:
    print (f"Something happened in ASI default_loading_callback {label}: {eee}\nAborting...")
    traceback = eee.__traceback__
    while traceback:
        print(f"{traceback.tb_frame.f_code.co_filename} : {traceback.tb_lineno}")
        traceback = traceback.tb_next
    MPI.COMM_WORLD.Abort(1)

class ASIlib:
  '''
    Python wrapper for dynamically loaded library with ASI API implementation
  '''
  def __init__(self, lib_file: str, initializer, mpi_comm=None, atoms=None, work_dir='asi.temp', logfile='asi.log'):
    '''
      Constructor for ASI library wrapper. Library itself is NOT loaded here.
      
      Parameters
      ----------
        lib_file : str 
          Path to the ASI-implementing shared object library
        initializer : function 
          Function or callable object that is supposed to create input files for the library in `work_dir`
        mpi_comm : int 
          MPI communicator, if is `None`, then `mpi4py.MPI.COMM_WORLD` will be used
        atoms : ase.Atoms
          ASE Atoms object for calculations. An internal copy will be created
        work_dir : str
          Working dir for `ASI_init()`, `ASI_run()`, and `ASI_finalize()` calls
        logfile : str 
          Log file for the ASI library
    '''
    self.lib_file = Path(lib_file).resolve()
    self.initializer = initializer
    if mpi_comm is not None:
      self.mpi_comm = mpi_comm
    else:
      from mpi4py import MPI
      self.mpi_comm = MPI.COMM_WORLD
    self.atoms = atoms.copy() if atoms is not None else None
    self.work_dir = Path(work_dir)
    self.work_dir.mkdir(parents=True, exist_ok=True)
    self.logfile = logfile

  def __enter__(self):
    return self.init()

  def __exit__(self, type, value, traceback):
    #Exception handling here
    #print ("__exit__: ", type, value, traceback)
    self.close()  

  def init(self):
    """
      Calls `self.initializer` (see __init__ argument of the same name), 
      load the ASI-implementing shared object library using `ctypes.CDLL`, and
      calls `ASI_init()`. All of the above is performed in `self.work_dir` as a current directory.
      
       No ASI calls are should be attempted before that function call.
    """
    curdir = os.getcwd()
    try:
      os.chdir(self.work_dir)
      
      if self.mpi_comm.Get_rank() == 0:
        self.initializer(self)
      self.mpi_comm.Barrier()
    
      # Load the FHI-aims library
      # mode=RTLD_GLOBAL is necessary to get rid of the error with MKL:
      # 		`INTEL MKL ERROR: /opt/intel/oneapi/mkl/2021.4.0/lib/intel64/libmkl_avx512.so.1: undefined symbol: mkl_sparse_optimize_bsr_trsm_i8.`
      # Details: https://bugs.launchpad.net/ubuntu/+source/intel-mkl/+bug/1947626
      self.lib = CDLL(self.lib_file, mode=RTLD_GLOBAL)
      self.scalapack = ScaLAPACK4py(self.lib)

      self.lib.ASI_n_atoms.restype = c_int
      self.lib.ASI_energy.restype = c_double
      self.lib.ASI_forces.restype = POINTER(c_double)
      if hasattr(self.lib, "ASI_stress"):
        self.lib.ASI_stress.restype = POINTER(c_double)
      self.lib.ASI_atomic_charges.restype = POINTER(c_double)
      self.lib.ASI_atomic_charges.argtypes  = [c_int,]
      self.lib.ASI_calc_esp.argtypes = [c_int, ndpointer(dtype=np.float64), ndpointer(dtype=np.float64), ndpointer(dtype=np.float64)]
      self.lib.ASI_register_dm_callback.argtypes = [dmhs_callback, c_void_p]
      self.lib.ASI_register_overlap_callback.argtypes = [dmhs_callback, c_void_p]
      self.lib.ASI_register_hamiltonian_callback.argtypes = [dmhs_callback, c_void_p]
      if hasattr(self.lib, "ASI_register_dm_init_callback"):
        self.lib.ASI_register_dm_init_callback.argtypes = [set_dmhs_callback, c_void_p]
      self.lib.ASI_register_external_potential.argtypes = [esp_callback, c_void_p];
      self.lib.ASI_is_hamiltonian_real.restype = c_bool
      self.lib.ASI_get_basis_size.restype = c_int
      if hasattr(self.lib, "ASI_get_basis_func_descr"):
        self.lib.ASI_get_basis_func_descr.restype = POINTER(ASI_basis_func_descr_generic)
        self.lib.ASI_get_basis_func_descr.argtypes = [c_int,]
      self.lib.ASI_get_nspin.restype = c_int
      self.lib.ASI_get_nkpts.restype = c_int
      self.lib.ASI_get_n_local_ks.restype = c_int
      self.lib.ASI_get_local_ks.restype = c_int
      self.lib.ASI_get_local_ks.argtypes = [ndpointer(dtype=np.int32),]
      if hasattr(self.lib, "ASI_is_scf_converged"):
        self.lib.ASI_is_scf_converged.restype = c_bool
      
      input_filename = {1:"dummy", 2:"dftb_in.hsd"}[self.lib.ASI_flavour()]
      self.lib.ASI_init(input_filename.encode('UTF-8'), self.logfile.encode('UTF-8'), c_int(self.mpi_comm.py2f()))
      if (self.lib.ASI_flavour() == 2):
        self.set_geometry() # DFTB+ ignores geometry from input files if used via API
      return self
    except Exception as err:
      print("Some error in ASIlib.init:", err)
      print ("traceback:\n", traceback.format_exc())
    finally:
      os.chdir(curdir)
  
  def close(self):
    '''
      Calls `ASI_finalize()`. No ASI calls are should be attempted after that function call.
    '''
    if not hasattr(self, 'lib'):
      return # allready closed

    curdir = os.getcwd()
    try:
      os.chdir(self.work_dir)
      self.lib.ASI_finalize()
      handle = self.lib._handle
      del self.lib
      if self.mpi_comm.Get_rank() == 0:
        os.system(f"cat {self.logfile} >> total.log")
    finally:
      os.chdir(curdir)
    
  def run(self):
    """
      Run calculation
      
      Calls `ASI_run()`
    """
    curdir = os.getcwd()
    try:
      os.chdir(self.work_dir)
      self.lib.ASI_run()
    except Exception as err:
      print(f"Exception, {err}")
    finally:
      os.chdir(curdir)

  def register_DM_init(self, dm_init_callback, dm_init_aux):
    """
      Register callback function to be called on Density Matrix initilaization before SCF loop
      
      Calls `ASI_register_dm_init_callback()`

      Parameters
      ----------
      dm_init_callback : dmhs_callback
        Callback function
      dm_init_aux : Object
        Auxiliary object for the callback
    """
    self.dm_init_callback = set_dmhs_callback(dm_init_callback)
    self.dm_init_aux = dm_init_aux
    self.lib.ASI_register_dm_init_callback(self.dm_init_callback, c_void_p.from_buffer(py_object(self.dm_init_aux)))

  def register_overlap_callback(self, overlap_callback, overlap_aux):
    """
      Register callback function to be called on overlap matrix calculation
      
      Calls `ASI_register_overlap_callback()`

      Parameters
      ----------
      overlap_callback : dmhs_callback
        Callback function
      overlap_aux : Object
        Auxiliary object for the callback
    """
    self.overlap_callback = dmhs_callback(overlap_callback)
    self.overlap_aux = overlap_aux
    self.lib.ASI_register_overlap_callback(self.overlap_callback, c_void_p.from_buffer(py_object(self.overlap_aux)))

  def register_hamiltonian_callback(self, hamiltonian_callback, hamiltonian_aux):
    """
      Register callback function to be called on hamiltonian matrix calculation
      
      Calls `ASI_register_hamiltonian_callback()`

      Parameters
      ----------
      hamiltonian_callback : dmhs_callback
        Callback function
      hamiltonian_aux : Object
        Auxiliary object for the callback
    """
    self.hamiltonian_callback = dmhs_callback(hamiltonian_callback)
    self.hamiltonian_aux = hamiltonian_aux
    self.lib.ASI_register_hamiltonian_callback(self.hamiltonian_callback, c_void_p.from_buffer(py_object(self.hamiltonian_aux)))

  def register_dm_callback(self, dm_callback, dm_aux):
    """
      Register callback function to be called on Density Matrix calculation
      
      Calls `ASI_register_dm_callback()`

      Parameters
      ----------
      dm_callback : dmhs_callback
        Callback function
      dm_aux : Object
        Auxiliary object for the callback
    """
    self.dm_callback = dmhs_callback(dm_callback)
    self.dm_aux = dm_aux
    self.lib.ASI_register_dm_callback(self.dm_callback, c_void_p.from_buffer(py_object(self.dm_aux)))

  def register_external_potential(self, ext_pot_func, ext_pot_aux_obj):
    """
      Register callback function for evaliation of external electrostatic potential
      
      Calls `ASI_register_external_potential()`

      Parameters
      ----------
      ext_pot_func : esp_callback
        Callback function
      ext_pot_aux_obj : Object
        Auxiliary object for the callback
    """
    self.ext_pot_func = esp_callback(ext_pot_func)
    self.ext_pot_aux_obj = ext_pot_aux_obj
    self.lib.ASI_register_external_potential(self.ext_pot_func, c_void_p.from_buffer(py_object(self.ext_pot_aux_obj)))

  def calc_esp(self, coords):
    """
      Compute electrostatic potential (ESP) and its gradient in arbitrary points
      
      Calls `ASI_calc_esp()`

      Parameters
      ----------
      coords : c_double[n, 3]
        Coordinates of points to compute ESP and its gradient
      
      Returns
      -------
      esp : c_double[n]
        ESP in corresponding points
      esp_grad : c_double[n, 3]
        ESP gradient in corresponding points
    """
    n = len(coords)
    esp = np.zeros((n,), dtype=c_double)
    esp_grad = np.zeros((n,3), dtype=c_double)
    self.lib.ASI_calc_esp(c_int(n), coords.ravel(), esp, esp_grad) 
    return esp, esp_grad

  @property
  def flavour(self):
    """
      int: ID of ASI implementation flavour
      
      Calls `ASI_flavour()`
    """
    return self.lib.ASI_flavour()
  
  @property
  def implementation(self):
    """
      str: Name of ASI implementation
      
      Calls `ASI_flavour()`
    """
    return {1:"FHI-AIMS", 2:"DFTB+"}[self.flavour]

  @property
  def n_atoms(self):
    """
      int: Number of atoms of the system

      Calls `ASI_n_atoms()`
    """
    return self.lib.ASI_n_atoms()

  @property
  def n_basis(self):
    """
      int: Number of basis functions

      Calls `ASI_get_basis_size()`
    """
    return self.lib.ASI_get_basis_size()
    
  @property
  def matrix_shape(self):
    '''
      int: Expected shape of H, S, or density matrices.
      
      (n_basis, n_basis) if matrices are real. (n_basis, n_basis, 2) if matrices are complex - last dimension is for Re/Im parts
    '''
    return (self.n_basis,self.n_basis) if self.is_hamiltonian_real else (self.n_basis,self.n_basis, 2) # resulting shape
  
  def get_basis_func_descr(self, bf_index):
    if not hasattr(self.lib, "ASI_get_basis_func_descr"):
      raise NotImplementedError("ASI_get_basis_func_descr not implemented in {self.lib_file}")
    
    descr_ptr = self.lib.ASI_get_basis_func_descr(bf_index)
    descr_ptr_specific = cast(descr_ptr, POINTER(basis_func_descr_types[descr_ptr.contents.basis_func_descr_type]))
    return descr_ptr_specific.contents

  @property
  def basis_atoms(self):
    """
      int[n_basis]: Atomic indices for each basis function

      Calls `ASI_get_basis_func_descr()` and `ASI_get_basis_size()`
    """
    return np.array([self.get_basis_func_descr(i).atom_index for i in range(self.n_basis)])

  @property
  def n_spin(self):
    """
      int: Number of spin channels

      Calls `ASI_get_nspin()`
    """
    return self.lib.ASI_get_nspin()

  @property
  def n_kpts(self):
    """
      int: Number of k-points

      Calls `ASI_get_nkpts()`
    """
    return self.lib.ASI_get_nkpts()

  @property
  def n_local_ks(self):
    """
      int: Number of pairs (k-point, spin-chanel-index) processed by current MPI process

      Calls `ASI_get_n_local_ks()`
    """
    return self.lib.ASI_get_n_local_ks()

  @property
  def local_ks(self):
    """
      int[n_local_ks * 2]: List of pairs (k-point, spin-chanel-index) processed by current MPI process

      Calls `ASI_get_local_ks()`
    """
    n = self.n_local_ks
    res = np.zeros((n*2,), dtype=c_int32)
    n2 =  self.lib.ASI_get_local_ks(res)
    assert n == n2
    return res

  @property
  def is_hamiltonian_real(self):
    """
      bool: `True`  if Hamiltonian of current system is real. `False` if Hamiltonian of current system is complex.
      
      Calls `ASI_is_hamiltonian_real()`
    """
    return self.lib.ASI_is_hamiltonian_real()

  @property
  def is_scf_converged(self):
    """
      bool: `True`  if SCF loop is converged.
      
      Calls `ASI_is_scf_converged()`
    """
    return self.lib.ASI_is_scf_converged()

  @property
  def total_forces(self):
    """
      c_double[n_atoms, 3 ]: Total forces acting on system atoms.
      
      Calls `ASI_forces()`
    """
    forces_ptr = self.lib.ASI_forces()
    if forces_ptr:
      return np.ctypeslib.as_array(forces_ptr, shape=(self.n_atoms, 3))
    else:
      return None

  @property
  def stress(self):
    """
      c_double[3, 3 ]: Stress tensor of the periodic system
      
      Calls `ASI_stress()`
    """
    stress_ptr = self.lib.ASI_stress()
    if stress_ptr:
      return np.ctypeslib.as_array(stress_ptr, shape=(3, 3))
    else:
      return None

  @property
  def atomic_charges(self):
    """
      c_double[n_atoms]: Atomic charges. Default partitioning scheme is implementation-dependent
      
      Calls `ASI_atomic_charges(-1)` (`-1` for default partitioning scheme)
    """
    chg_ptr = self.lib.ASI_atomic_charges(-1)
    if chg_ptr:
      return np.ctypeslib.as_array(chg_ptr, shape=(self.n_atoms,)).copy()
    else:
      return None

  @property
  def total_energy(self):
    """
      c_double: Total energy of the system
      
      Calls `ASI_energy()`
    """
    return self.lib.ASI_energy()
  
  def set_geometry(self):
    coords_ptr = (self.atoms.positions / units.Bohr).ctypes.data_as(c_void_p)
    if any(self.atoms.pbc):
      lattice_ptr = (self.atoms.cell.ravel() / units.Bohr).ctypes.data_as(c_void_p)
      self.lib.ASI_set_geometry(coords_ptr, len(self.atoms), lattice_ptr)
    else:
      self.lib.ASI_set_atom_coords(coords_ptr, len(self.atoms))

  @property
  def keep_density_matrix(self):
    """
      bool : Flag to save Density Matrix in `self.dm_storage` dict and count number
        of the matrix calculations in `self.dm_calc_cnt` dict.
        The flag may be set to `True` to save only the last SCF interation matrix,
        or to `"history"` to save matrices for each SCF iteration.
        
        Dictionaries are indexed by (k-point, spin-chanel-index) pairs.
    """
    return hasattr(self, 'dm_callback')

  @keep_density_matrix.setter
  def keep_density_matrix(self, value):
    assert value, 'callback unsetting not implemented'
    if self.keep_density_matrix:
      self.dm_storage.clear()
      self.dm_calc_cnt.clear()
      return

    self.dm_storage = {}
    self.dm_calc_cnt = {}
    self.register_dm_callback(default_saving_callback, (self, self.dm_storage, self.dm_calc_cnt, value=='history', 'DM calc'))

  @property
  def keep_hamiltonian(self):
    """
      bool : Flag to save Hamiltonian matrix in `self.hamiltonian_storage` dict and count number
        of the matrix calculations in `self.hamiltonian_calc_cnt` dict.
        The flag may be set to `True` to save only the last SCF interation matrix,
        or to `"history"` to save matrices for each SCF iteration.
        
        Dictionaries are indexed by (k-point, spin-chanel-index) pairs.
    """
    return hasattr(self, 'hamiltonian_callback')

  @keep_hamiltonian.setter
  def keep_hamiltonian(self, value):
    assert value, 'callback unsetting not implemented'
    if self.keep_hamiltonian:
      self.hamiltonian_storage.clear()
      self.hamiltonian_calc_cnt.clear()
      return

    self.hamiltonian_storage = {}
    self.hamiltonian_calc_cnt = {}
    self.register_hamiltonian_callback(default_saving_callback, (self, self.hamiltonian_storage, self.hamiltonian_calc_cnt, value=='history', 'H calc'))

  @property
  def keep_overlap(self):
    """
      bool : Flag to save overlap matrix in `self.overlap_storage` dict and count number
        of the matrix calculations in `self.overlap_calc_cnt` dict.

        Dictionaries are indexed by (k-point, spin-chanel-index) pairs.
    """
    return hasattr(self, 'overlap_callback')

  @keep_overlap.setter
  def keep_overlap(self, value):
    assert value, 'callback unsetting not implemented'
    if self.keep_overlap:
      self.overlap_storage.clear()
      self.overlap_calc_cnt.clear()
      return

    self.overlap_storage = {}
    self.overlap_calc_cnt = {}
    self.register_overlap_callback(default_saving_callback, (self, self.overlap_storage, self.overlap_calc_cnt, False, 'S calc'))

  @property
  def init_density_matrix(self):
    """
      bool / c_double[n_basis, n_basis] : Set with a density matrix t be used for SCF loop initialization.
      Reading that property returns True if the density matrix initialization is enabled
    """
    return hasattr(self, 'dm_init_callback')
 
  @init_density_matrix.setter
  def init_density_matrix(self, value):
    self.dm_init_storage = value
    self.register_DM_init(default_loading_callback, (self, self.dm_init_storage, 'DM init'))




from ase.calculators.calculator import Calculator, all_changes, PropertyNotImplementedError
from .pyasi import ASIlib
from ase import units
from ase.parallel import parprint
from ase.stress import full_3x3_to_voigt_6_stress
"""
# FHI-AIMS conversion constants
bohr    = 0.52917721
hartree = 27.2113845
hartree_over_bohr = 51.42206426
bohr_over_hartree = 0.019446905
"""
bohr    = units.Bohr
hartree = units.Hartree


class ASI_ASE_calculator(Calculator):
  '''
    ASI ASI calc
  '''
  implemented_properties = ['energy', 'free_energy', 'forces', 'charges', 'stress']
  supported_changes = {'positions', 'cell'}

  def __init__(self, asi_lib, init_func=None, mpi_comm=None, atoms=None, work_dir='asi.temp', logfile='asi.log'):
    try:
      Calculator.__init__(self)
      #!! self.atoms = atoms.copy()
      if isinstance(asi_lib, ASIlib):
        self.asi = asi_lib
      else:
        self.asi = ASIlib(asi_lib, init_func, mpi_comm, atoms, work_dir, logfile)
        self.asi.init()
      self.DM_init_callback = None
    except Exception  as err:
      print ("Error in init",err)
      raise err

  def todict(self):
      d = {'type': 'calculator',
           'name': 'ASI wrapper'}
      return d

  def calculate(self, atoms=None, properties=['energy'],
                system_changes=supported_changes):
      bad = [change for change in system_changes
             if change not in self.supported_changes] # TODO now is ignored
      
      #parprint("calculate", properties, system_changes, bad, atoms, self.atoms, self.asi.atoms)

      # First time calculate() is called, system_changes will be
      # all_changes.  After that, only positions and cell may change.
      if self.atoms is not None and any(bad):
          raise PropertyNotImplementedError(
              'Cannot change {} through ASI API.  '
              .format(bad if len(bad) > 1 else bad[0]))

      if len(system_changes) > 0:
        self.atoms = atoms.copy()
        self.asi.atoms = atoms
        self.asi.set_geometry()
        self.asi.run()

      results = {}
      results['free_energy'] = results['energy'] = self.asi.total_energy * hartree
      
      if 'forces' in properties:
        if self.asi.total_forces is not None:
          results['forces'] = self.asi.total_forces * (hartree / bohr)

      if 'stress' in properties:
        if self.asi.stress is not None:
          stress_3x3 = self.asi.stress * (hartree / (bohr**3))
          results['stress'] = full_3x3_to_voigt_6_stress(stress_3x3)

      # Charges computation breaks total_energy on subsequent SCF calculations in AIMS:  results['charges'] = self.asi.atomic_charges

      self.results.update(results)

  def close(self):
      self.asi.close()

  def __del__(self):
      self.close()



from ctypes import cdll, CDLL, RTLD_GLOBAL
from ctypes import POINTER, byref, c_int, c_int64, c_int32, c_bool, c_char_p, c_double, c_void_p, CFUNCTYPE, py_object, cast, byref, Structure
import ctypes

from scipy.linalg import block_diag
from ase.data import chemical_symbols
from ase.geometry import get_distances
import ase, os, warnings
import numpy as np
import numpy.typing as npt
from typing import List, Any
from scipy.io import netcdf_file
from asi4py.pyasi import matrix_asi_to_numpy
from scipy.optimize import LinearConstraint, minimize


def free_energy_fun(x, E, beta):
  '''
    Free energy
  '''
  return np.sum(x * (E*beta - np.log(x)))

def free_energy_jac(x, E, beta):
  '''
    Free energy gradient
  '''
  return E*beta - np.log(x)-1

def get_elems_ordered(atoms):
  '''
    List elements in the order of tiers.
    https://gitlab.com/ase/ase/-/commit/7b70eddd026154d636faf404cc2f8c7b08d89667
    https://mail.python.org/pipermail/python-dev/2017-December/151283.html
  '''
  return list(dict.fromkeys(atoms.symbols))

def split_dm(dm, magmom:float, iS:int):
  '''
    Split spin-paired density matrix into up/down spin-polarizes DM
  '''
  w,v = np.linalg.eigh(dm)
  w /= 2
  w_positive_indices = np.where(w>0)[0]
  E = -np.log(w[w_positive_indices])
  n_avg = np.sum(w)
  n_up   = n_avg + magmom/2
  n_down = n_avg - magmom/2
  b_eq = n_up if iS==1 else n_down
  w_new = minimize(lambda x: free_energy_fun(x, E, 0.1) , w[w_positive_indices], jac=lambda x: free_energy_jac(x, E, 0.1), bounds=[(1e-10, 1.0) for _ in E], constraints=[LinearConstraint(np.ones((1, len(E))), b_eq, b_eq), ] ).x
   
  w[w_positive_indices] = w_new
  return v @ np.diag(w) @ v.T

def build_dm_free_atoms(atoms, elem_dms, n_spin:int, iS:int):
  '''
    Build atoms density matrix from single-atomic matrices for each element
    :param n_spin: number of spin channels. 1 or 2.
    :param iS: index of the spin channel. 1 or 2.
  '''
  assert n_spin in [1,2], n_spin
  
  atomic_dms = [elem_dms[s] for s in atoms.symbols]
  if n_spin==2:
    atomic_dms = [split_dm(dm,magmom,iS) for dm,magmom in zip(atomic_dms, atoms.get_initial_magnetic_moments())]

  return block_diag(*atomic_dms).T

def load_atomic_dm(elem:str, path:str=None):
  if path is None:
    path = os.environ['ASI_FREE_ATOM_DMS']
  elemZ = chemical_symbols.index(elem)
  with netcdf_file(f'{path}/{elemZ:03d}_{elem}.ncdf', mmap=False) as f:
    return f.variables['DM'].data

def save_atomic_dm(dm, elem:str, path:str=None):
  '''
    Save density matrix for element `elem` into file `{path}/{elemZ:03d}_{elem}.ncdf`
    If `path` is `None` then it defaults to environment variable `ASI_FREE_ATOM_DMS`
    This files maybe used by :class:`PredictFreeAtoms` predictor
  '''
  if path is None:
    path = os.environ['ASI_FREE_ATOM_DMS']
  elemZ = chemical_symbols.index(elem)
  with netcdf_file(f'{path}/{elemZ:03d}_{elem}.ncdf','w', mmap=False) as f:
    f.createDimension('n_basis', dm.shape[0])
    f.createVariable('DM', dm.dtype, ('n_basis', 'n_basis'))[:,:] = dm

def bool2int_selector(atoms_selector):
  atoms_selector = np.array(atoms_selector)
  if np.issubdtype(atoms_selector.dtype, bool):
    atoms_selector = np.where(atoms_selector)[0]
  assert np.issubdtype(atoms_selector.dtype, np.integer)
  return atoms_selector

class PredictDMByAtoms:
  '''
    This is a base class for density matrix predictors. It is meant to be subclassed.
    Subclasses must overwrite the method :func:`PredictDMByAtoms.__call__` that takes :class:`ase.Atoms` object as argument
    and returns :class:`numpy.ndarray` with a density matrix guess.
  '''
  def __init__(self):
    pass
  
  def register_DM_init(self, asi):
    '''
      Register this density matrix predictor as an ASI callback for density matrix initialization
      
      :param asi: an instance of :class:`asi4py.pyasi.ASILib`
      
    '''
    self.asi = asi
    asi.register_DM_init(PredictDMByAtoms.dm_init_callback, self)

  def dm_init_callback(self, iK, iS, blacs_descr, data, matrix_descr_ptr):
    '''
      ASI DM init callback. Not to be invoked directly.
    '''
    self = cast(self, py_object).value
    n_basis = self.asi.n_basis
    m = self(self.asi.atoms, iK, iS) if self.asi.scalapack.is_root(blacs_descr) else None
    
    assert m is None or (m.shape == (n_basis, n_basis)), \
                     f"m.shape=={m.shape} != n_basis=={n_basis}"
    self.asi.scalapack.scatter_numpy(m, blacs_descr, data)
    return 1

  def __call__(self, atoms, iK:int=1, iS:int=1):
    '''
      This method is meant to be overwritten by derived classes.
    '''
    raise RuntimeError("Not implemented in base class")
    #return build_dm_free_atoms(atoms, self.elem_dms)

class PredictFreeAtoms(PredictDMByAtoms):
  '''
    Density matrix predictor that uses single-atomic density matrices for initialization.
    See :class:`PredictFreeAtoms.__init__` for details
  '''
  def __init__(self, elem_dms:dict[str, Any]=None, elem_dm_path:str=None,  n_spin:int=1):
    '''
      If ``elem_dms`` parameter is not ``None``, then ``elem_dm_path`` should be ``None`.
      If both ``elem_dms`` and ``elem_dm_path`` are ``None``, then  ``elem_dm_path` value
      defaults to the ``ASI_FREE_ATOM_DMS`` environment variable.
      
      :param elem_dms: a dictionary that maps chemical elements to its density matrices
      :param elem_dm_path: a path to a folder that contains `*.npz` files with density matrices of chemical elements
      :param n_spin:  number of spin channels, should be 2 if initial magnetic moment is specified
    '''
    super().__init__()
    self.n_spin = n_spin
    assert (elem_dms is None) or (elem_dm_path is None)
    self.elem_dms = elem_dms
    if self.elem_dms is None:
      self.elem_dm_path = elem_dm_path if elem_dm_path is not None else os.environ['ASI_FREE_ATOM_DMS']

  def __call__(self, atoms, iK:int=1, iS:int=1):
    if iK != 1:
      return np.zeros((self.asi.n_basis, self.asi.n_basis))
    if self.elem_dms is None:
      assert self.elem_dm_path is not None
      self.elem_dms = {elem:load_atomic_dm(elem, self.elem_dm_path) for elem in get_elems_ordered(atoms)}
    return build_dm_free_atoms(atoms, self.elem_dms, self.n_spin, iS)

class PredictConstAtoms(PredictDMByAtoms):
  '''
    Density matrix predictor that returns a density matrix passed to its constructor, independently from
    atomic coordinates.
  '''
  def __init__(self, const_atoms:ase.Atoms, const_dm:Any):
    '''
      :param const_atoms: :class:`ase.Atoms` object that is only used for checking order of chemical elements on density matrix prediction
      :param const_dm: a density matrix that will be returned as initial guess on SCF loop initialization
    '''
    super().__init__()
    self.const_atoms = const_atoms
    self.const_dm = const_dm

  def __call__(self, atoms, iK:int=1, iS:int=1):
    np.testing.assert_allclose(atoms.numbers, self.const_atoms.numbers)
    
    np.testing.assert_allclose(
      atoms.positions - atoms.get_center_of_mass(), 
      self.const_atoms.positions - self.const_atoms.get_center_of_mass(), 
      atol=1e-1, rtol=1e-1)
    
    return self.const_dm

def select_basis_indices(all_basis_atoms, atoms_indices):
  '''
    Returns indices of basis functions for selected atoms.
    
    :param all_basis_atoms: atomic indices for all basis functions
    :param atoms_indices: indices of selected atoms
  '''
  return np.where(np.any(all_basis_atoms[None,:] == atoms_indices[:, None], axis=0))[0]

class PredictFrankensteinDM(PredictDMByAtoms):
  '''
    Density matrix predictor that "stitches" full density matrix prediction from 
    predictions of multiples predictors of smaller, possibly overlapping subsystems of the full system.
  '''
  def __init__(self, predictors_and_selectors:list[tuple[PredictDMByAtoms, list[int]]]):
    '''
      :param predictors_and_selectors: list of pairs. The first pair element is callable, that 
        should return predicted density matrix for a subsystem; it can be derived from :class:`PredictDMByAtoms`.
        The second pair element is the list of atomc that constitute the subsystem.
        
    '''
    # unzip https://stackoverflow.com/a/12974504/3213940
    self.predictors, atoms_selectors = list(zip(*predictors_and_selectors))
    self.atoms_groups_indices = list(map(bool2int_selector, atoms_selectors))

    if True: # extended assertion check
      all_selected_atoms_set = set().union(*self.atoms_groups_indices)
      max_range_atoms_set = set(range(max(all_selected_atoms_set) + 1))
      missed_atoms = max_range_atoms_set - all_selected_atoms_set
      # Heuristic check: doesn't guarantie full atoms coverage, because actual
      # number of atoms is not known here and max(all_selected_atoms_set) is 
      # just an heuristic
      assert len(missed_atoms) == 0, f"Missed atoms: {missed_atoms}"
  
  def register_DM_init(self, asi):
    super().register_DM_init(asi)
    self.init_basis_indices(asi)

  def init_basis_indices(self, asi):
    '''
      Loads indices of basis functions for subsystems from an ASI library.
      It should only be called if one is going to predict density matrix without registering 
      this predictor via :func:`PredictDMByAtoms.register_DM_init`
    '''
    all_basis_atoms = asi.basis_atoms
    self.basis_indices = [select_basis_indices(all_basis_atoms, atoms_group) for atoms_group in self.atoms_groups_indices]
    self.n_basis = asi.n_basis

    if True:
      all_selected_basis_indices = set().union(*self.basis_indices)
      total_basis_set = set(range(self.n_basis))
      missed_basis_functions = total_basis_set - all_selected_basis_indices
      assert len(missed_basis_functions) == 0, f'Basis functions missed from selection: {missed_basis_functions}'

  def __call__(self, atoms, iK:int=1, iS:int=1):
    assert len(self.predictors) == len(self.atoms_groups_indices)
    assert len(self.predictors) == len(self.basis_indices)

    total_dm = np.zeros((self.n_basis, self.n_basis), dtype=np.float64)
    total_dm_cnt = np.zeros(total_dm.shape, dtype=int)
    for predictor, atoms_group_indices, basis_group_indices in zip(self.predictors, self.atoms_groups_indices, self.basis_indices):
      total_dm[basis_group_indices[np.newaxis,:], basis_group_indices[:, np.newaxis]] += predictor(atoms[atoms_group_indices])
      total_dm_cnt[basis_group_indices[np.newaxis,:], basis_group_indices[:, np.newaxis]] += 1
    
    assert (total_dm[total_dm_cnt==0]==0).all()
    total_dm_cnt[total_dm_cnt==0] = 1 # to avoid division by zero

    return (total_dm / total_dm_cnt).T

class PredictSeqDM(PredictDMByAtoms):
  '''
    Extrapolates density matrix using kernel ridge regression from a few previous geometries.
    This predictor uses another predictor as a baseline and extrapolates only deviations of 
    baseline predictors from ground-state density matrices.
    The method :func:`PredictSeqDM.update_errs` should be invoked after SCF convergence to provide
    the predictor with ground-state density matrix, for extrapolation model adjustment.
    
  '''
  class PredictionSequence:
    def __init__(self):
      self.preds = []
      self.descrs = []
      self.errs = []

    def predict_err(self, predicted_dm, descr):
      from sklearn.gaussian_process.kernels import RBF
      from sklearn.kernel_ridge import KernelRidge
      k = len(self.preds)
      assert k == len(self.descrs), f'{k=} {len(self.descrs)=}'
      assert k == len(self.errs), f'{k=} {len(self.errs)=}'
      if k == 0:
        return np.zeros(predicted_dm.shape)

      X = np.hstack([np.array(self.preds).reshape((k, -1)), np.array(self.descrs).reshape((k, -1))])
      x = np.hstack([predicted_dm.ravel(), descr.ravel()]).reshape((1, -1))
      Y = np.array(self.errs).reshape((k, -1))
      

      X_mean = np.mean(X, axis=0)
      Y_mean = np.mean(Y, axis=0)
      
      X -= X_mean
      Y -= Y_mean
      x -= X_mean
      #------------------

      with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Ill-conditioned matrix")
        reg = KernelRidge(alpha=1e-15, kernel='rbf')
        reg.fit(X, Y)
        y = reg.predict(x)

      #------------------
      y += Y_mean
      P = y.reshape(predicted_dm.shape)
      return P
      
  
  def __init__(self, base_predictor:PredictDMByAtoms, n_hist:int):
    '''
      :param base_predictor: Baseline predictor, for example :class:`PredictFreeAtoms` or :class:`PredictFrankensteinDM` 
      :param n_hist: history size for extrapolation
    '''
    self.base_predictor = base_predictor
    self.predseq = {} # dict (iK,iS):PredictionSequence
    self.n_hist = n_hist
  
  def __call__(self, atoms, iK:int=1, iS:int=1):
    predicted_dm = self.base_predictor(atoms, iK, iS)

    R,d = get_distances(atoms.positions)
    lowtri = np.tri(len(atoms), len(atoms), -1)==1
    invd = 1/d[lowtri]
    descr = invd

    if (iK, iS) not in self.predseq:
      self.predseq[(iK, iS)] = PredictSeqDM.PredictionSequence()

    predicted_err = self.predseq[(iK, iS)].predict_err(predicted_dm, descr)

    if len(self.predseq[(iK, iS)].preds) == self.n_hist:
      self.predseq[(iK, iS)].preds.pop(0)
      self.predseq[(iK, iS)].descrs.pop(0)
      self.predseq[(iK, iS)].errs.pop(0)
    self.predseq[(iK, iS)].preds.append(predicted_dm)
    self.predseq[(iK, iS)].descrs.append(descr)

    return (predicted_dm - predicted_err).T

  def register_DM_init(self, asi):
    '''
      Register this density matrix predictor as an ASI callback for density matrix initialization,
      and registers itself to process DM evaluation callback for errors update.
      
      :param asi: an instance of :class:`asi4py.pyasi.ASILib`
      
    '''
    super().register_DM_init(asi)
    asi.dm_calc_cnt = {}
    asi.dm_storage = {}
    asi.register_dm_callback(PredictSeqDM.dm_callback, self)

  def dm_callback(self, iK, iS, descr, data, matrix_descr_ptr):
    self = cast(self, py_object).value
    self.asi.dm_calc_cnt[(iK, iS)] = self.asi.dm_calc_cnt.get((iK, iS),0) + 1
    data = matrix_asi_to_numpy(self.asi, descr, data, matrix_descr_ptr)
    if data is not None:
      self.asi.dm_storage[(iK, iS)] = data.copy()
  
  def update_errs(self):
    '''
      This method should be called after SCF convergence, to adjust extrapolation model.
    '''
    for (iK, iS),pred_item in self.predseq.items():
      assert len(pred_item.preds) == len(pred_item.errs) + 1
      pred_item.errs.append(pred_item.preds[-1] - self.asi.dm_storage[(iK, iS)])
      self.asi.dm_calc_cnt[(iK, iS)] = 0
    
    self.register_DM_init(self.asi)
  







from .pyasi import ASIlib, triang2herm_inplace, triang_packed2full_hermit, HS_to_DM
from .asecalc import ASI_ASE_calculator

__all__ = ['ASI_ASE_calculator', 'ASIlib', 'triang2herm_inplace', 'triang_packed2full_hermit', 'HS_to_DM']
